"""Various utility functions."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple

import numpy as np
import torch
import torch.nn.functional as F
import yaml

import ptlflow
from ptlflow.utils.external.raft import InputPadder as _InputPadder


class InputPadder(_InputPadder):
    """Pads images such that dimensions are divisible by stride.

    This is just a wrapper for ptlflow.utils.external.raft.InputPadder.
    """

    def __init__(
        self,
        dims: Sequence[int],
        stride: int
    ) -> None:
        """Initialize InputPadder.

        Parameters
        ----------
        dims : Sequence[int]
            The shape of the original input. It must have at least two elements. It is assumed that the last two dimensions
            are (height, width).
        stride : int
            The number to compute the amount of padding. The padding will be applied so that the input size is divisible
            by stride.
        """
        super().__init__(dims, stride=stride)


class InputScaler(object):
    """Scale 2D torch.Tensor input to a target size, and then rescale it back to the original size."""

    def __init__(
        self,
        orig_shape: Tuple[int, int],
        size: Optional[Tuple[int, int]] = None,
        scale_factor: Optional[float] = 1.0,
        interpolation_mode: str = 'bilinear',
        interpolation_align_corners: bool = False
    ) -> None:
        """Initialize InputScaler.

        Parameters
        ----------
        orig_shape : Tuple[int, int]
            The shape of the input tensor before the scale. I.e., the shape to which it will be rescaled back.
        size : Optional[Tuple[int, int]], optional
            The desired size after scaling defined as (height, width). If not provided, then scale_factor will be used instead.
        scale_factor : Optional[float], default 1.0
            This value is only used if size is None. The multiplier that will be applied to the original shape to scale
            the input.
        interpolation_mode : str, default 'bilinear'
            How to perform the interpolation. It must be a value accepted by the 'mode' argument from
            torch.nn.functional.interpolate function.
        interpolation_align_corners : bool, default False
            Whether the interpolation keep the corners aligned. As defined in torch.nn.functional.interpolate.

        See Also
        --------
        torch.nn.functional.interpolate : The function used to scale the inputs.
        """
        super().__init__()
        self.orig_height, self.orig_width = orig_shape[-2:]
        if size is None:
            self.tgt_height = int(self.orig_height * scale_factor)
            self.tgt_width = int(self.orig_width * scale_factor)
        else:
            self.tgt_height, self.tgt_width = size

        self.interpolation_mode = interpolation_mode
        self.interpolation_align_corners = interpolation_align_corners

    def scale(
        self,
        x: torch.Tensor
    ) -> torch.Tensor:
        """Scale the input to the target size specified during initialization.

        Parameters
        ----------
        x : torch.Tensor
            The input to be scaled. Its shape must be (..., C, H, W), where ... means any number of dimensions.

        Returns
        -------
        torch.Tensor
            The scaled input.
        """
        return self._scale_keep_dims(x, (self.tgt_height, self.tgt_width))

    def unscale(
        self,
        x: torch.Tensor
    ) -> torch.Tensor:
        """Scale the input to back to the original size defined during initialization.

        Parameters
        ----------
        x : torch.Tensor
            The input to be rescaled back. Its shape must be (..., C, H, W), where ... means any number of dimensions.

        Returns
        -------
        torch.Tensor
            The rescaled input.
        """
        return self._scale_keep_dims(x, (self.orig_height, self.orig_width))

    def _scale_keep_dims(
        self,
        x: torch.Tensor,
        size: Tuple[int, int]
    ) -> torch.Tensor:
        """Scale the input to a given size while keeping the other dimensions intact.

        Parameters
        ----------
        x : torch.Tensor
            The input to be rescaled back. Its shape must be (..., C, H, W), where ... means any number of dimensions.
        size : Tuple[int, int]
            The target size to scale the input.

        Returns
        -------
        torch.Tensor
            The rescaled input.
        """
        x_shape = x.shape
        x = x.view(-1, x.shape[-3], x.shape[-2], x.shape[-1])
        x = F.interpolate(
            x, size=size, mode=self.interpolation_mode,
            align_corners=self.interpolation_align_corners)
        new_shape = list(x_shape)
        new_shape[-2], new_shape[-1] = x.shape[-2], x.shape[-1]
        x = x.view(new_shape)
        return x


def add_datasets_to_parser(
    parser: ArgumentParser,
    dataset_config_path: str
) -> ArgumentParser:
    """Add dataset paths as parser arguments.

    The dataset names and default paths are loaded from a yaml file.

    Parameters
    ----------
    parser : ArgumentParser
        An initialized parser, this function will add more arguments to it.
    dataset_config_path : str
        The path to the yaml file containing the dataset paths.

    Returns
    -------
    ArgumentParser
        The updated parser.
    """
    with open(dataset_config_path, 'r') as f:
        dataset_paths = yaml.safe_load(f)
    for name, path in dataset_paths.items():
        parser.add_argument(
            f'--{name}_root_dir', type=str, default=path,
            help=f'Path to the root of the {name} dataset')
    return parser


def config_logging() -> None:
    """Initialize logging parameters."""
    log_dir = Path('ptlflow_logs')
    log_dir.mkdir(exist_ok=True)
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO,
        handlers=[
            logging.FileHandler(log_dir / 'log_run.txt'),
            logging.StreamHandler()
        ]
    )


def count_parameters(model: torch.nn.Module) -> int:
    """Return the number of trainable parameters of a model.

    Taken from: https://discuss.pytorch.org/t/how-do-i-check-the-number-of-parameters-of-a-model/4325/7.

    Parameters
    ----------
    model : torch.nn.Module
        The model to count the parameters from.

    Returns
    -------
    int
        The number of trainable parameters of the given model.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def get_list_of_available_models_list() -> List[str]:
    """Return a list of the names of the available models.

    Returns
    -------
    list[str]
        The list with the model names.
    """
    return sorted(ptlflow.models_dict.keys())


def make_divisible(
    v: int,
    div: int
) -> int:
    """Decrease a number v until it is divisible by div.

    Parameters
    ----------
    v : int
        The number to be made divisible.
    div : int
        The divisor value.

    Returns
    -------
    int
        The new value of v which is divisible by div.
    """
    return max(div, v - (v % div))


def release_gpu(
    tensors_dict: Dict[str, Any]
) -> Dict[str, Any]:
    """Detach and move to cpu the tensors from the input dict.

    The non-tensor elements are kept intact.

    Parameters
    ----------
    tensors_dict : Dict[str, Any]
        A dictionary containing the tensors to move.

    Returns
    -------
    Dict[str, Any]
        The same dictionary, but with the tensors moved to cpu.
    """
    for k, v in tensors_dict.items():
        if isinstance(v, torch.Tensor):
            tensors_dict[k] = v.detach().cpu()
            del v
    return tensors_dict


def tensor_dict_to_numpy(
    tensor_dict: Dict[str, torch.Tensor],
    padder: Optional[InputPadder] = None
) -> Dict[str, np.ndarray]:
    """Convert all tensors into numpy format, changing the shape from CHW to HWC.

    If "flows" is available, then a color representation "flows_viz" is added to the outputs.

    Parameters
    ----------
    tensor_dict : dict[str, torch.Tensor]
        A dictionary with the torch.Tensor inputs/outputs of the model.
    padder: InputPadder
        Helper to unpad the images back to their original sizes.

    Returns
    -------
    dict[str, np.ndarray]
        The torch.Tensor entries from tensor_dict converted to numpy format.
    """
    npy_dict = {k: v.detach().cpu() for k, v in tensor_dict.items() if isinstance(v, torch.Tensor)}
    if padder is not None:
        npy_dict = {k: padder.unpad(v) for k, v in npy_dict.items()}
    for k, v in npy_dict.items():
        while len(v.shape) > 3:
            v = v[0]
        npy_dict[k] = v
    npy_dict = {k: v.permute(1, 2, 0).numpy() for k, v in npy_dict.items()}
    return npy_dict



from argparse import ArgumentParser, Namespace
from typing import Any, Dict, List, Optional

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn
import torch.nn.functional as F

from .warp import WarpingLayer
from ..base_model.base_model import BaseModel


class FeatureExtractor(nn.Module):
    def __init__(self):
        super(FeatureExtractor, self).__init__()

        leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        self.convs = nn.ModuleList([
            nn.Sequential(
                nn.Conv2d(3, 32, 7, 1, 3),
                leaky_relu
            ),
            nn.Sequential(
                nn.Conv2d(32, 32, 3, 2, 1),
                leaky_relu,
                nn.Conv2d(32, 32, 3, 1, 1),
                leaky_relu,
                nn.Conv2d(32, 32, 3, 1, 1),
                leaky_relu
            ),
            nn.Sequential(
                nn.Conv2d(32, 64, 3, 2, 1),
                leaky_relu,
                nn.Conv2d(64, 64, 3, 1, 1),
                leaky_relu
            ),
            nn.Sequential(
                nn.Conv2d(64, 96, 3, 2, 1),
                leaky_relu,
                nn.Conv2d(96, 96, 3, 1, 1),
                leaky_relu
            ),
            nn.Sequential(
                nn.Conv2d(96, 128, 3, 2, 1),
                leaky_relu
            ),
            nn.Sequential(
                nn.Conv2d(128, 192, 3, 2, 1),
                leaky_relu
            )
        ])

    def forward(
        self,
        images: torch.Tensor
    ) -> List[torch.Tensor]:
        features = []

        x = images.view(-1, *images.shape[2:])
        for i, conv in enumerate(self.convs):
            x = conv(x)
            if i > 1:
                features.append(x.view(*images.shape[:2], *x.shape[1:]))

        return features[::-1]


class Matching(nn.Module):
    def __init__(
        self,
        level: int,
        num_levels: int = 4,
        div_flow: float = 20.0
    ) -> None:
        super(Matching, self).__init__()

        corr_stride = [1, 1, 1, 2][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2**(num_levels-i+1) for i in range(num_levels)][level]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        if level == 0:
            self.up_flow = None
        else:
            self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)

        if level < 3:
            self.up_corr = None
        else:
            self.up_corr = nn.ConvTranspose2d(49, 49, 4, 2, 1, bias=False, groups=49)

        self.flow_net = nn.Sequential(
            nn.Conv2d(49, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 96, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(96, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(32, 2, flow_kernel_size, 1, flow_kernel_size//2)
        )

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=7, padding=0, stride=corr_stride, dilation_patch=corr_stride)

        self.warp = WarpingLayer()

    def forward(
        self,
        feats: torch.Tensor,
        flow: Optional[torch.Tensor]
    ) -> torch.Tensor:
        warped_feat2 = feats[:, 1]
        if flow is not None:
            flow = self.up_flow(flow)
            warped_feat2 = self.warp(feats[:, 1], flow, feats.shape[-2], feats.shape[-1], 1.0/self.mult)

        corr = self.leaky_relu(self.corr(feats[:, 0], warped_feat2))
        corr = corr.view(corr.shape[0], -1, corr.shape[3], corr.shape[4])
        corr = corr / feats.shape[2]
        if self.up_corr is not None:
            corr = self.up_corr(corr)
        new_flow = self.flow_net(corr)
        if flow is not None:
            new_flow = flow + new_flow
        return new_flow


class SubPixel(nn.Module):
    def __init__(
        self,
        level: int,
        num_levels: int = 4,
        div_flow: float = 20.0
    ) -> None:
        super(SubPixel, self).__init__()

        inputs_dims = [386, 258, 194, 130][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2**(num_levels-i+1) for i in range(num_levels)][level]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        self.feat_net = nn.Sequential(
            nn.Conv2d(inputs_dims, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 96, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(96, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu
        )

        self.flow_net = nn.Conv2d(32, 2, flow_kernel_size, 1, flow_kernel_size//2)
        
        self.warp = WarpingLayer()

    def forward(
        self,
        feats: torch.Tensor,
        flow: torch.Tensor
    ) -> torch.Tensor:
        feat_warped = self.warp(feats[:, 1], flow, feats.shape[-2], feats.shape[-1], 1.0/self.mult)
        x = torch.cat([feats[:, 0], feat_warped, flow], dim=1)
        x = self.feat_net(x)
        new_flow = self.flow_net(x)
        new_flow = flow + new_flow
        return new_flow, x


class Regularization(nn.Module):
    def __init__(
        self,
        level: int,
        num_levels: int = 4,
        div_flow: float = 20.0
    ) -> None:
        super(Regularization, self).__init__()

        inputs_dims = [195, 131, 99, 67][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2**(num_levels-i+1) for i in range(num_levels)][level]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        if level < 2:
            self.feat_conv = nn.Sequential()
        else:
            self.feat_conv = nn.Sequential(
                nn.Conv2d(inputs_dims-3, 128, 1, 1, 0),
                self.leaky_relu
            )
            inputs_dims = 131

        self.feat_net = nn.Sequential(
            nn.Conv2d(inputs_dims, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(32, 32, 3, 1, 1),
            self.leaky_relu,
        )

        if level < 2:
            self.dist = nn.Conv2d(32, flow_kernel_size**2, 3, 1, 1)
        else:
            self.dist = nn.Sequential(
                nn.Conv2d(32, flow_kernel_size**2, (flow_kernel_size, 1), 1, (flow_kernel_size//2, 0)),
                nn.Conv2d(flow_kernel_size**2, flow_kernel_size**2, (1, flow_kernel_size), 1, (0, flow_kernel_size//2))
            )

        self.unfold = nn.Unfold(flow_kernel_size, padding=flow_kernel_size//2)

        self.warp = WarpingLayer()

    def forward(
        self,
        images: torch.Tensor,
        feats: torch.Tensor,
        flow: torch.Tensor
    ) -> torch.Tensor:
        img2_warped = self.warp(images[:, 1], flow, images.shape[-2], images.shape[-1], 1.0/self.mult)
        img_diff_norm = torch.norm(images[:, 0] - img2_warped[:, 1], p=2, dim=1, keepdim=True)

        flow_mean = flow.view(*flow.shape[:2], -1).mean(dim=-1)[..., None, None]
        flow_nomean = flow - flow_mean
        feat = self.feat_conv(feats[:, 0])
        x = torch.cat([img_diff_norm, flow_nomean, feat], dim=1)
        x = self.feat_net(x)
        dist = self.dist(x)
        dist = dist.square().neg()
        dist = (dist - dist.max(dim=1, keepdim=True)[0]).exp()
        div = dist.sum(dim=1, keepdim=True)

        reshaped_flow_x = self.unfold(flow[:, :1])
        reshaped_flow_x = reshaped_flow_x.view(*reshaped_flow_x.shape[:2], *flow.shape[2:4])
        flow_smooth_x = (reshaped_flow_x * dist).sum(dim=1, keepdim=True) / div

        reshaped_flow_y = self.unfold(flow[:, 1:2])
        reshaped_flow_y = reshaped_flow_y.view(*reshaped_flow_y.shape[:2], *flow.shape[2:4])
        flow_smooth_y = (reshaped_flow_y * dist).sum(dim=1, keepdim=True) / div

        flow = torch.cat([flow_smooth_x, flow_smooth_y], dim=1)

        return flow, x


class PseudoSubpixel(nn.Module):
    def __init__(
        self
    ) -> None:
        super(PseudoSubpixel, self).__init__()

        self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)

        self.flow_net = nn.Sequential(
            nn.ConvTranspose2d(32, 32, 4, 2, 1),
            nn.Conv2d(32, 2, 7, 1, 3)
        )

    def forward(
        self,
        sub_feat: torch.Tensor,
        flow: torch.Tensor
    ) -> torch.Tensor:
        return self.up_flow(flow) + self.flow_net(sub_feat)


class PseudoRegularization(nn.Module):
    def __init__(
        self
    ) -> None:
        super(PseudoRegularization, self).__init__()
        
        self.feat_net = nn.Sequential(
            nn.ConvTranspose2d(32, 32, 4, 2, 1),
            nn.Conv2d(32, 49, (7, 1), 1, (3, 0)),
            nn.Conv2d(49, 49, (1, 7), 1, (0, 3))
        )

        self.unfold = nn.Unfold(7, padding=3)

    def forward(
        self,
        reg_feat: torch.Tensor,
        flow: torch.Tensor
    ) -> torch.Tensor:
        dist = self.feat_net(reg_feat)
        dist = dist.square().neg()
        dist = (dist - dist.max(dim=1, keepdim=True)[0]).exp()
        div = dist.sum(dim=1, keepdim=True)

        reshaped_flow_x = self.unfold(flow[:, :1])
        reshaped_flow_x = reshaped_flow_x.view(*reshaped_flow_x.shape[:2], *flow.shape[2:4])
        flow_smooth_x = (reshaped_flow_x * dist).sum(dim=1, keepdim=True) / div

        reshaped_flow_y = self.unfold(flow[:, 1:2])
        reshaped_flow_y = reshaped_flow_y.view(*reshaped_flow_y.shape[:2], *flow.shape[2:4])
        flow_smooth_y = (reshaped_flow_y * dist).sum(dim=1, keepdim=True) / div

        flow = torch.cat([flow_smooth_x, flow_smooth_y], dim=1)

        return flow


class LiteFlowNet2(BaseModel):
    pretrained_checkpoints = {
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/liteflownet2-sintel-1e1eb282.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        super(LiteFlowNet2, self).__init__(
            args=args,
            loss_fn=None,
            output_stride=32)

        self.num_levels = 4

        self.feature_net = FeatureExtractor()
        self.matching_nets = nn.ModuleList([Matching(i, self.num_levels, self.args.div_flow) for i in range(self.num_levels)])
        self.subpixel_nets = nn.ModuleList([SubPixel(i, self.num_levels, self.args.div_flow) for i in range(self.num_levels)])
        self.regularization_nets = nn.ModuleList([Regularization(i, self.num_levels, self.args.div_flow) for i in range(self.num_levels)])

        if self.args.use_pseudo_regularization:
            self.pseudo_subpixel = PseudoSubpixel()
            self.pseudo_regularization = PseudoRegularization()
            self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)
        else:
            self.up_flow = nn.ConvTranspose2d(2, 2, 8, 4, 2, bias=False, groups=2)

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=20.0)
        parser.add_argument('--use_pseudo_regularization', action='store_true')
        return parser

    def forward(
        self,
        inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        images = inputs['images']
        images_mean = images.view(*images.shape[:3], -1).mean(dim=-1)[..., None, None]
        images = images - images_mean
        
        feats_pyr = self.feature_net(images)
        images_pyr = self._create_images_pyr(images, feats_pyr)
        
        flow_preds = []
        flow = None

        for i in range(self.num_levels):
            flow = self.matching_nets[i](feats_pyr[i], flow)
            flow, sub_feat = self.subpixel_nets[i](feats_pyr[i], flow)
            flow, reg_feat = self.regularization_nets[i](images_pyr[i], feats_pyr[i], flow)
            flow_preds.append(flow)
        
        if self.args.use_pseudo_regularization:
            flow = self.pseudo_subpixel(sub_feat, flow)
            flow = self.pseudo_regularization(reg_feat, flow)
            flow = self.up_flow(flow)
        else:
            flow = self.up_flow(flow)
        flow = flow * self.args.div_flow

        outputs = {}
        if self.training:
            outputs['flow_preds'] = flow_preds
            outputs['flows'] = flow[:, None]
        else:
            outputs['flows'] = flow[:, None]
        return outputs


    def _create_images_pyr(
        self,
        images: torch.Tensor,
        feats_pyr: List[torch.Tensor]
    ) -> List[torch.Tensor]:
        batch_size = images.shape[0]
        images = images.view(-1, *images.shape[2:]).detach()
        images_pyr = [
            F.interpolate(images, size=feats_pyr[i].shape[-2:], mode='bilinear', align_corners=False)
            for i in range(len(feats_pyr))]
        images_pyr = [im.view(batch_size, -1, *im.shape[1:]) for im in images_pyr]
        return images_pyr


class LiteFlowNet2PseudoReg(LiteFlowNet2):
    pretrained_checkpoints = {
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/liteflownet2-kitti-da069fca.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        args.use_pseudo_regularization = True
        super(LiteFlowNet2PseudoReg, self).__init__(args=args)



'''
Portions of this code copyright 2017, Clement Pinard
'''
from argparse import Namespace
from pathlib import Path

import torch
import torch.nn as nn
from torch.nn import init

from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetSD(FlowNetBase):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetsd-things-c5f3124e.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        super(FlowNetSD, self).__init__(args)

        self.conv0   = conv(self.args.batch_norm,  6,   64)
        self.conv1   = conv(self.args.batch_norm,  64,   64, stride=2)
        self.conv1_1 = conv(self.args.batch_norm,  64,   128)
        self.conv2   = conv(self.args.batch_norm,  128,  128, stride=2)
        self.conv2_1 = conv(self.args.batch_norm,  128,  128)
        self.conv3   = conv(self.args.batch_norm, 128,  256, stride=2)
        self.conv3_1 = conv(self.args.batch_norm, 256,  256)
        self.conv4   = conv(self.args.batch_norm, 256,  512, stride=2)
        self.conv4_1 = conv(self.args.batch_norm, 512,  512)
        self.conv5   = conv(self.args.batch_norm, 512,  512, stride=2)
        self.conv5_1 = conv(self.args.batch_norm, 512,  512)
        self.conv6   = conv(self.args.batch_norm, 512, 1024, stride=2)
        self.conv6_1 = conv(self.args.batch_norm,1024, 1024)

        self.deconv5 = deconv(1024,512)
        self.deconv4 = deconv(1026,256)
        self.deconv3 = deconv(770,128)
        self.deconv2 = deconv(386,64)

        self.inter_conv5 = i_conv(self.args.batch_norm,  1026,   512)
        self.inter_conv4 = i_conv(self.args.batch_norm,  770,   256)
        self.inter_conv3 = i_conv(self.args.batch_norm,  386,   128)
        self.inter_conv2 = i_conv(self.args.batch_norm,  194,   64)

        self.predict_flow6 = predict_flow(1024)
        self.predict_flow5 = predict_flow(512)
        self.predict_flow4 = predict_flow(256)
        self.predict_flow3 = predict_flow(128)
        self.predict_flow2 = predict_flow(64)

        self.upsampled_flow6_to_5 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow5_to_4 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow4_to_3 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow3_to_2 = nn.ConvTranspose2d(2, 2, 4, 2, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear', align_corners=False)

    def forward(self, inputs):
        x = inputs['images']

        x = x.view(x.shape[0], x.shape[1]*x.shape[2], x.shape[3], x.shape[4])
        out_conv0 = self.conv0(x)
        out_conv1 = self.conv1_1(self.conv1(out_conv0))
        out_conv2 = self.conv2_1(self.conv2(out_conv1))

        out_conv3 = self.conv3_1(self.conv3(out_conv2))
        out_conv4 = self.conv4_1(self.conv4(out_conv3))
        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)
        
        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)
        out_interconv5 = self.inter_conv5(concat5)
        flow5       = self.predict_flow5(out_interconv5)

        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)
        out_interconv4 = self.inter_conv4(concat4)
        flow4       = self.predict_flow4(out_interconv4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        
        concat3 = torch.cat((out_conv3,out_deconv3,flow4_up),1)
        out_interconv3 = self.inter_conv3(concat3)
        flow3       = self.predict_flow3(out_interconv3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)

        concat2 = torch.cat((out_conv2,out_deconv2,flow3_up),1)
        out_interconv2 = self.inter_conv2(concat2)
        flow2 = self.predict_flow2(out_interconv2)

        outputs = {}

        if self.training:
            outputs['flow_preds'] = [flow2.float(),flow3.float(),flow4.float(),flow5.float(),flow6.float()]
            outputs['flows'] = self.upsample1(flow2.float())[:, None] / self.args.div_flow
        else:
            outputs['flows'] = self.upsample1(flow2.float())[:, None] / self.args.div_flow

        return outputs



"""This code is mostly taken and adapted from flowpy.

https://gitlab-research.centralesupelec.fr/2018seznecm/flowpy
"""

#
# Original license below
#
# MIT License
#
# Copyright (c) 2020 Univ. Paris-Saclay, CNRS, CentraleSupelec,
#                    Thales Research & Technology
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np
from pathlib import Path
from warnings import warn
import png
import struct
from collections import namedtuple
from itertools import accumulate

DEFAULT_TRANSITIONS = (15, 6, 4, 11, 13, 6)


def flow_to_rgb(flow, flow_max_radius=None, background="bright", custom_colorwheel=None):
    """
    Creates a RGB representation of an optical flow.

    Parameters
    ----------
    flow: numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] should be the x-displacement
        flow[..., 1] should be the y-displacement

    flow_max_radius: float, optional
        Set the radius that gives the maximum color intensity, useful for comparing different flows.
        Default: The normalization is based on the input flow maximum radius.

    background: str, optional
        States if zero-valued flow should look 'bright' or 'dark'
        Default: "bright"

    custom_colorwheel: numpy.ndarray
        Use a custom colorwheel for specific hue transition lengths.
        By default, the default transition lengths are used.

    Returns
    -------
    rgb_image: numpy.ndarray
        A 2D RGB image that represents the flow

    Raises
    ------
    ValueError
        If the background is invalid.

    See Also
    --------
    make_colorwheel

    """
    valid_backgrounds = ("bright", "dark")
    if background not in valid_backgrounds:
        raise ValueError(f"background should be one the following: {valid_backgrounds}, not {background}")

    wheel = make_colorwheel() if custom_colorwheel is None else custom_colorwheel

    complex_flow = flow[..., 0] + 1j * flow[..., 1]
    complex_flow, nan_mask = replace_nans(complex_flow)

    radius, angle = np.abs(complex_flow), np.angle(complex_flow)

    if flow_max_radius is None:
        flow_max_radius = np.max(radius)

    if flow_max_radius > 0:
        radius /= flow_max_radius

    ncols = len(wheel)

    # Map the angles from (-pi, pi] to [0, 2pi) to [0, ncols - 1)
    angle[angle < 0] += 2 * np.pi
    angle = angle * ((ncols - 1) / (2 * np.pi))

    # Make the wheel cyclic for interpolation
    wheel = np.vstack((wheel, wheel[0]))

    # Interpolate the hues
    (angle_fractional, angle_floor), angle_ceil = np.modf(angle), np.ceil(angle)
    angle_fractional = angle_fractional.reshape((angle_fractional.shape) + (1,))
    float_hue = (wheel[angle_floor.astype(np.int32)] * (1 - angle_fractional)
                 + wheel[angle_ceil.astype(np.int32)] * angle_fractional)

    ColorizationArgs = namedtuple("ColorizationArgs", [
        'move_hue_valid_radius',
        'move_hue_oversized_radius',
        'invalid_color'])

    def move_hue_on_V_axis(hues, factors):
        return hues * np.expand_dims(factors, -1)

    def move_hue_on_S_axis(hues, factors):
        return 255. - np.expand_dims(factors, -1) * (255. - hues)

    if background == "dark":
        parameters = ColorizationArgs(move_hue_on_V_axis, move_hue_on_S_axis,
                                      np.array([255, 255, 255], dtype=np.float32))
    else:
        parameters = ColorizationArgs(move_hue_on_S_axis, move_hue_on_V_axis,
                                      np.array([0, 0, 0], dtype=np.float32))

    colors = parameters.move_hue_valid_radius(float_hue, radius)

    oversized_radius_mask = radius > 1
    colors[oversized_radius_mask] = parameters.move_hue_oversized_radius(
        float_hue[oversized_radius_mask],
        1 / radius[oversized_radius_mask]
    )
    colors[nan_mask] = parameters.invalid_color

    return colors.astype(np.uint8)


def make_colorwheel(transitions=DEFAULT_TRANSITIONS):
    """
    Creates a color wheel.

    A color wheel defines the transitions between the six primary hues:
    Red(255, 0, 0), Yellow(255, 255, 0), Green(0, 255, 0), Cyan(0, 255, 255), Blue(0, 0, 255) and Magenta(255, 0, 255).

    Parameters
    ----------
    transitions: sequence_like
        Contains the length of the six transitions.
        Defaults to (15, 6, 4, 11, 13, 6), based on humain perception.

    Returns
    -------
    colorwheel: numpy.ndarray
        The RGB values of the transitions in the color space.

    Notes
    -----
    For more information, take a look at
    https://web.archive.org/web/20051107102013/http://members.shaw.ca/quadibloc/other/colint.htm

    """

    colorwheel_length = sum(transitions)

    # The red hue is repeated to make the color wheel cyclic
    base_hues = map(np.array,
                    ([255, 0, 0], [255, 255, 0], [0, 255, 0],
                     [0, 255, 255], [0, 0, 255], [255, 0, 255],
                     [255, 0, 0]))

    colorwheel = np.zeros((colorwheel_length, 3), dtype="uint8")
    hue_from = next(base_hues)
    start_index = 0
    for hue_to, end_index in zip(base_hues, accumulate(transitions)):
        transition_length = end_index - start_index

        colorwheel[start_index:end_index] = np.linspace(
            hue_from, hue_to, transition_length, endpoint=False)
        hue_from = hue_to
        start_index = end_index

    return colorwheel


def replace_nans(array, value=0):
    nan_mask = np.isnan(array)
    array[nan_mask] = value

    return array, nan_mask


def flow_write(output_file, flow, format=None):
    """
    Writes optical flow to file.

    Parameters
    ----------
    output_file: {str, pathlib.Path, file}
        Path of the file to write or file object.
    flow: numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] should be the x-displacement
        flow[..., 1] should be the y-displacement
    format: str, optional
        Specify in what format the flow is written, accepted formats: "png" or "flo"
        If None, it is guessed on the file extension

    See Also
    --------
    flow_read

    """

    output_format = guess_extension(output_file, override=format)

    with FileManager(output_file, "wb") as f:
        if output_format == "png":
            flow_write_png(f, flow)
        else:
            flow_write_flo(f, flow)


def flow_read(input_file, format=None):
    """
    Reads optical flow from file

    Parameters
    ----------
    input_file: {str, pathlib.Path, file}
        Path of the file to read or file object.
    format: str, optional
        Specify in what format the flow is read, accepted formats: "png" or "flo"
        If None, it is guess on the file extension

    Returns
    -------
    flow: numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] is the x-displacement
        flow[..., 1] is the y-displacement

    Notes
    -----

    The flo format is dedicated to optical flow and was first used in Middlebury optical flow database.
    The original definition can be found here: http://vision.middlebury.edu/flow/code/flow-code/flowIO.cpp

    The png format uses 16-bit RGB png to store optical flows.
    It was developped along with the KITTI Vision Benchmark Suite.
    More information can be found here: http://www.cvlibs.net/datasets/kitti/eval_scene_flow.php?benchmark=flow

    The both handle flow with invalid ``invalid'' values, to deal with occlusion for example.
    We convert such invalid values to NaN.

    See Also
    --------
    flow_write

    """

    input_format = guess_extension(input_file, override=format)

    with FileManager(input_file, "rb") as f:
        if input_format == "png":
            output = flow_read_png(f)
        else:
            output = flow_read_flo(f)

    return output


def flow_read_flo(f):
    if (f.read(4) != b'PIEH'):
        warn(f"{f.name} does not have a .flo file signature")

    width, height = struct.unpack("II", f.read(8))
    result = np.fromfile(f, dtype="float32").reshape((height, width, 2))

    # Set invalid flows to NaN
    mask_u = np.greater(np.abs(result[..., 0]), 1e9, where=(~np.isnan(result[..., 0])))
    mask_v = np.greater(np.abs(result[..., 1]), 1e9, where=(~np.isnan(result[..., 1])))

    result[mask_u | mask_v] = np.NaN

    return result


def flow_write_flo(f, flow):
    SENTINEL = 1666666800.0  # Only here to look like Middlebury original files
    height, width, _ = flow.shape

    image = flow.copy()
    image[np.isnan(image)] = SENTINEL

    f.write(b'PIEH')
    f.write(struct.pack("II", width, height))
    image.astype(np.float32).tofile(f)


def flow_read_png(f):
    width, height, stream, *_ = png.Reader(f).read()

    file_content = np.concatenate(list(stream)).reshape((height, width, 3))
    flow, valid = file_content[..., 0:2], file_content[..., 2]

    flow = (flow.astype(np.float32) - 2 ** 15) / 64.

    flow[~valid.astype(bool)] = np.NaN

    return flow


def flow_write_png(f, flow):
    SENTINEL = 0.  # Only here to look like original KITTI files
    height, width, _ = flow.shape
    flow_copy = flow.copy()

    valid = ~(np.isnan(flow[..., 0]) | np.isnan(flow[..., 1]))
    flow_copy[~valid] = SENTINEL

    flow_copy = (flow_copy * 64. + 2 ** 15).astype(np.uint16)
    image = np.dstack((flow_copy, valid))

    writer = png.Writer(width, height, bitdepth=16, greyscale=False)
    writer.write(f, image.reshape((height, 3 * width)))


class FileManager:
    def __init__(self, abstract_file, mode):
        self.abstract_file = abstract_file
        self.opened_file = None
        self.mode = mode

    def __enter__(self):
        if isinstance(self.abstract_file, str):
            self.opened_file = open(self.abstract_file, self.mode)
        elif isinstance(self.abstract_file, Path):
            self.opened_file = self.abstract_file.open(self.mode)
        else:
            return self.abstract_file

        return self.opened_file

    def __exit__(self, exc_type, exc_value, traceback):
        if self.opened_file is not None:
            self.opened_file.close()


def guess_extension(abstract_file, override=None):
    if override is not None:
        return override

    if isinstance(abstract_file, str):
        return Path(abstract_file).suffix[1:]
    elif isinstance(abstract_file, Path):
        return abstract_file.suffix[1:]

    return Path(abstract_file.name).suffix[1:]



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
import torch.nn.functional as tf
import logging


def conv(in_planes, out_planes, kernel_size=3, stride=1, dilation=1, isReLU=True):
    if isReLU:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, dilation=dilation,
                      padding=((kernel_size - 1) * dilation) // 2, bias=True),
            nn.LeakyReLU(0.1, inplace=True)
        )
    else:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, dilation=dilation,
                      padding=((kernel_size - 1) * dilation) // 2, bias=True)
        )


def initialize_msra(modules):
    logging.info("Initializing MSRA")
    for layer in modules:
        if isinstance(layer, nn.Conv2d):
            nn.init.kaiming_normal_(layer.weight)
            if layer.bias is not None:
                nn.init.constant_(layer.bias, 0)

        elif isinstance(layer, nn.ConvTranspose2d):
            nn.init.kaiming_normal_(layer.weight)
            if layer.bias is not None:
                nn.init.constant_(layer.bias, 0)

        elif isinstance(layer, nn.LeakyReLU):
            pass

        elif isinstance(layer, nn.Sequential):
            pass


def upsample2d_as(inputs, target_as, mode="bilinear"):
    _, _, h, w = target_as.size()
    return tf.interpolate(inputs, [h, w], mode=mode, align_corners=True)


def rescale_flow(flow, div_flow, width_im, height_im, to_local=True):
    if to_local:
        u_scale = float(flow.size(3) / width_im / div_flow)
        v_scale = float(flow.size(2) / height_im / div_flow)
    else:
        u_scale = float(width_im * div_flow / flow.size(3))
        v_scale = float(height_im * div_flow / flow.size(2))

    u, v = flow.chunk(2, dim=1)
    u = u * u_scale
    v = v * v_scale

    return torch.cat([u, v], dim=1)


class FeatureExtractor(nn.Module):
    def __init__(self, num_chs):
        super(FeatureExtractor, self).__init__()
        self.num_chs = num_chs
        self.convs = nn.ModuleList()

        for l, (ch_in, ch_out) in enumerate(zip(num_chs[:-1], num_chs[1:])):
            layer = nn.Sequential(
                conv(ch_in, ch_out, stride=2),
                conv(ch_out, ch_out)
            )
            self.convs.append(layer)

    def forward(self, x):
        feature_pyramid = []
        for conv in self.convs:
            x = conv(x)
            feature_pyramid.append(x)

        return feature_pyramid[::-1]


def get_grid(x):
    grid_H = torch.linspace(-1.0, 1.0, x.size(3)).view(1, 1, 1, x.size(3)).expand(x.size(0), 1, x.size(2), x.size(3))
    grid_V = torch.linspace(-1.0, 1.0, x.size(2)).view(1, 1, x.size(2), 1).expand(x.size(0), 1, x.size(2), x.size(3))
    grid = torch.cat([grid_H, grid_V], 1)
    grids_cuda = grid.requires_grad_(False).to(dtype=x.dtype, device=x.device)
    return grids_cuda


class WarpingLayer(nn.Module):
    def __init__(self):
        super(WarpingLayer, self).__init__()

    def forward(self, x, flow, height_im, width_im, div_flow, cuda=True, debug=False):
        if debug:
            print('in warping')
            print(x.mean())
            print(flow.mean())

        flo_list = []
        flo_w = flow[:, 0] * 2 / max(width_im - 1, 1) / div_flow
        flo_h = flow[:, 1] * 2 / max(height_im - 1, 1) / div_flow
        flo_list.append(flo_w)
        flo_list.append(flo_h)
        flow_for_grid = torch.stack(flo_list).transpose(0, 1)
        if debug:
            print(flow_for_grid.mean())

        grid = torch.add(get_grid(x), flow_for_grid).transpose(1, 2).transpose(2, 3)
        if not cuda:
            x = x.cpu()
            grid = grid.cpu()
        x_warp = tf.grid_sample(x, grid, align_corners=True)
        if debug:
            print(x_warp.mean())

        mask = torch.ones_like(x, requires_grad=False)
        if not cuda:
            mask = mask.cpu()
            grid = grid.cpu()

        if debug:
            print(mask.mean())

        mask = tf.grid_sample(mask.float(), grid.float(), align_corners=True)
        if debug:
            print(mask.mean())

        mask[mask < 0.99999] = 0.
        mask[mask > 0] = 1.

        if debug:
            print(mask.mean())

        warped = x_warp * mask.float()
        if debug:
            print(warped.mean())
            print(warped.shape)
        return warped.to(device=x.device)


class OpticalFlowEstimator(nn.Module):
    def __init__(self, ch_in):
        super(OpticalFlowEstimator, self).__init__()

        self.convs = nn.Sequential(
            conv(ch_in, 128),
            conv(128, 128),
            conv(128, 96),
            conv(96, 64),
            conv(64, 32)
        )
        self.conv_last = conv(32, 2, isReLU=False)

    def forward(self, x):
        x_intm = self.convs(x)
        return x_intm, self.conv_last(x_intm)


class FlowEstimatorDense(nn.Module):
    def __init__(self, ch_in):
        super(FlowEstimatorDense, self).__init__()
        self.conv1 = conv(ch_in, 128)
        self.conv2 = conv(ch_in + 128, 128)
        self.conv3 = conv(ch_in + 256, 96)
        self.conv4 = conv(ch_in + 352, 64)
        self.conv5 = conv(ch_in + 416, 32)
        self.conv_last = conv(ch_in + 448, 2, isReLU=False)

    def forward(self, x):
        x1 = torch.cat([self.conv1(x), x], dim=1)
        x2 = torch.cat([self.conv2(x1), x1], dim=1)
        x3 = torch.cat([self.conv3(x2), x2], dim=1)
        x4 = torch.cat([self.conv4(x3), x3], dim=1)
        x5 = torch.cat([self.conv5(x4), x4], dim=1)
        x_out = self.conv_last(x5)
        return x5, x_out


class OcclusionEstimator(nn.Module):
    def __init__(self, ch_in):
        super(OcclusionEstimator, self).__init__()
        self.convs = nn.Sequential(
            conv(ch_in, 128),
            conv(128, 128),
            conv(128, 96),
            conv(96, 64),
            conv(64, 32)
        )
        self.conv_last = conv(32, 1, isReLU=False)

    def forward(self, x):
        x_intm = self.convs(x)
        return x_intm, self.conv_last(x_intm)


class OccEstimatorDense(nn.Module):
    def __init__(self, ch_in):
        super(OccEstimatorDense, self).__init__()
        self.conv1 = conv(ch_in, 128)
        self.conv2 = conv(ch_in + 128, 128)
        self.conv3 = conv(ch_in + 256, 96)
        self.conv4 = conv(ch_in + 352, 64)
        self.conv5 = conv(ch_in + 416, 32)
        self.conv_last = conv(ch_in + 448, 1, isReLU=False)

    def forward(self, x):
        x1 = torch.cat([self.conv1(x), x], dim=1)
        x2 = torch.cat([self.conv2(x1), x1], dim=1)
        x3 = torch.cat([self.conv3(x2), x2], dim=1)
        x4 = torch.cat([self.conv4(x3), x3], dim=1)
        x5 = torch.cat([self.conv5(x4), x4], dim=1)
        x_out = self.conv_last(x5)
        return x5, x_out


class ContextNetwork(nn.Module):
    def __init__(self, ch_in):
        super(ContextNetwork, self).__init__()

        self.convs = nn.Sequential(
            conv(ch_in, 128, 3, 1, 1),
            conv(128, 128, 3, 1, 2),
            conv(128, 128, 3, 1, 4),
            conv(128, 96, 3, 1, 8),
            conv(96, 64, 3, 1, 16),
            conv(64, 32, 3, 1, 1),
            conv(32, 2, isReLU=False)
        )

    def forward(self, x):
        return self.convs(x)


class OccContextNetwork(nn.Module):
    def __init__(self, ch_in):
        super(OccContextNetwork, self).__init__()

        self.convs = nn.Sequential(
            conv(ch_in, 128, 3, 1, 1),
            conv(128, 128, 3, 1, 2),
            conv(128, 128, 3, 1, 4),
            conv(128, 96, 3, 1, 8),
            conv(96, 64, 3, 1, 16),
            conv(64, 32, 3, 1, 1),
            conv(32, 1, isReLU=False)
        )

    def forward(self, x):
        return self.convs(x)



from argparse import ArgumentParser, Namespace

from torch.functional import norm

from ..base_model.base_model import BaseModel
from .losses import MultiScale


class FlowNetBase(BaseModel):
    def __init__(self,
                 args: Namespace):
        super(FlowNetBase, self).__init__(
            args=args,
            loss_fn=MultiScale(
                startScale=args.loss_start_scale,
                numScales=args.loss_num_scales,
                l_weight=args.loss_base_weight,
                norm=args.loss_norm
            ),
            output_stride=64)

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=20.0)
        parser.add_argument('--input_channels', type=int, default=6)
        parser.add_argument('--batch_norm', action='store_true')
        parser.add_argument('--loss_start_scale', type=float, default=4)
        parser.add_argument('--loss_num_scales', type=int, default=5)
        parser.add_argument('--loss_base_weight', type=float, default=0.32)
        parser.add_argument('--loss_norm', type=str, default='L2')
        return parser



from argparse import ArgumentParser, Namespace
from pathlib import Path

import numpy as np
try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn
import torch.nn.functional as F

from .hd3_ops import *
from .dla_up import DLAEncoder as dlaup_encoder
from .vgg import VGGEncoder as vgg_encoder
from . import decoder as PreDecoder
from .hd3losses import LossCalculator
from ..base_model.base_model import BaseModel

BatchNorm = nn.BatchNorm2d


class Context(nn.Module):

    def __init__(self, inplane, classes):
        super(Context, self).__init__()
        self.num_convs = 7
        ch = [inplane, 128, 128, 128, 128, 128, 128, 128]
        dilations = [1, 1, 2, 4, 8, 16, 1]
        for i in range(self.num_convs):
            setattr(
                self, 'dc_conv_{}'.format(i),
                nn.Sequential(
                    nn.Conv2d(
                        ch[i],
                        ch[i + 1],
                        kernel_size=3,
                        stride=1,
                        padding=dilations[i],
                        dilation=dilations[i],
                        bias=False), BatchNorm(ch[i + 1]),
                    nn.ReLU(inplace=True)))
        self.cls = nn.Conv2d(
            ch[-1], classes, kernel_size=1, stride=1, padding=0, bias=True)

    def forward(self, x):
        out = x
        for i in range(self.num_convs):
            dc_conv = getattr(self, 'dc_conv_' + str(i))
            out = dc_conv(out)
        out = self.cls(out)
        return out, None


class Decoder(nn.Module):

    def __init__(self, inplane, block, classes, up_classes):
        super(Decoder, self).__init__()
        self.mapping = block(inplane, 128)
        self.cls = nn.Sequential(
            BatchNorm(128), nn.ReLU(inplace=True),
            nn.Conv2d(
                128, classes, kernel_size=1, stride=1, padding=0, bias=True))
        self.up = None
        if up_classes > 0:
            self.up = nn.Sequential(
                BatchNorm(128), nn.ReLU(inplace=True),
                nn.ConvTranspose2d(
                    128,
                    up_classes,
                    kernel_size=4,
                    stride=2,
                    padding=1,
                    bias=False), BatchNorm(up_classes), nn.ReLU(inplace=True))

    def forward(self, x):
        out = self.mapping(x)
        prob = self.cls(out)
        up_feat = self.up(out) if self.up else None
        return prob, up_feat


class BaseHD3(BaseModel):
    def __init__(self, args, context):
        super(BaseHD3, self).__init__(
            args, LossCalculator(args.task), int(2**args.downsample))

        self.context = context

        if self.args.task == 'flow':
            self.args.corr_range = self.args.corr_range[:5]

        self.task = self.args.task
        self.dim = 1 if self.args.task == 'stereo' else 2
        self.levels = len(self.args.corr_range)
        if self.task == 'flow':
            self.classes = [(2 * d + 1)**2 for d in self.args.corr_range]
        else:
            self.classes = [2 * d + 1 for d in self.args.corr_range]

        if self.args.encoder == 'vgg':
            pyr_channels = [16, 32, 64, 96, 128, 196]
            assert self.levels <= len(pyr_channels)
            self.encoder = vgg_encoder(pyr_channels)
        elif self.args.encoder == 'dlaup':
            pyr_channels = [16, 32, 64, 128, 256, 512, 512]
            self.encoder = dlaup_encoder(pyr_channels)
        else:
            raise ValueError('Unknown encoder {}'.format(self.args.encoder))

        if self.args.decoder == 'resnet':
            dec_block = PreDecoder.ResnetDecoder
        elif self.args.decoder == 'hda':
            dec_block = PreDecoder.HDADecoder
        else:
            raise ValueError('Unknown decoder {}'.format(self.args.decoder))

        feat_d_offset = pyr_channels[::-1]
        feat_d_offset[0] = 0
        up_d_offset = [0] + self.classes[1:]

        for l in range(self.levels):
            setattr(self, 'cost_bn_{}'.format(l), BatchNorm(self.classes[l]))
            input_d = self.classes[l] + feat_d_offset[l] + up_d_offset[
                l] + self.dim * (
                    l > 0)
            if l < self.levels - 1:
                up_classes = self.classes[l + 1]
            else:
                up_classes = -1
            if self.context and l == self.levels - 1:
                setattr(self, 'Decoder_{}'.format(l),
                        Context(input_d, self.classes[l]))
            else:
                setattr(
                    self, 'Decoder_{}'.format(l),
                    Decoder(
                        input_d,
                        dec_block,
                        self.classes[l],
                        up_classes=up_classes))

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=9, padding=0)

        for m in self.modules():
            classname = m.__class__.__name__
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                m.weight.data.normal_(0, 0.02)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif classname.find('BatchNorm') != -1:
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--task', type=str, default='flow', choices=['flow', 'stereo'])
        parser.add_argument('--encoder', type=str, default='dlaup', choices=['vgg', 'dlaup'])
        parser.add_argument('--decoder', type=str, default='hda', choices=['hda', 'resnet'])
        parser.add_argument('--downsample', type=int, default=6)
        parser.add_argument('--corr_range', type=int, nargs='+', default=[4, 4, 4, 4, 4, 4])
        return parser

    def shift(self, x, vect):
        if vect.size(1) < 2:
            vect = disp2flow(vect)
        return flow_warp(x, vect)

    def forward(self, inputs):
        x = inputs['images']

        mean = torch.from_numpy(np.array([0.485, 0.456, 0.406])).to(dtype=x.dtype, device=x.device)[None, None, :, None, None]
        std = torch.from_numpy(np.array([0.229, 0.224, 0.225])).to(dtype=x.dtype, device=x.device)[None, None, :, None, None]
        x = (x - mean) / std

        # extract pyramid features
        bs = x.size(0)
        feat_list = self.encoder(
            torch.cat([x[:, 0], x[:, 1]], 0))
        fp_0 = [f[:bs, :, :, :] for f in feat_list[::-1]]
        fp_1 = [f[bs:, :, :, :] for f in feat_list[::-1]]

        ms_pred = []
        for l in range(self.levels):
            ref_feat = fp_0[l]
            tar_feat = fp_1[l]

            if l == 0:
                tar_feat_corr = tar_feat
            else:
                tar_feat_corr = self.shift(tar_feat, up_curr_vect)

            cost_vol = self.corr(ref_feat, tar_feat_corr)
            cost_vol = cost_vol.view(cost_vol.shape[0], -1, cost_vol.shape[3], cost_vol.shape[4])
            cost_vol = cost_vol / ref_feat.shape[1]
            if self.task == 'stereo':
                c = self.classes[l] // 2
                cost_vol = cost_vol[:, c * (2 * c + 1):(c + 1) *
                                    (2 * c + 1), :, :].contiguous()
            cost_bn = getattr(self, 'cost_bn_' + str(l))
            cost_vol = cost_bn(cost_vol)

            if l == 0:
                decoder_input = cost_vol
            else:
                decoder_input = torch.cat(
                    [cost_vol, ref_feat, ms_pred[-1][-1], up_curr_vect], 1)

            decoder = getattr(self, 'Decoder_' + str(l))
            prob_map, up_feat = decoder(decoder_input)

            curr_vect = density2vector(prob_map, self.dim, True)
            if l > 0:
                curr_vect += up_curr_vect
            if self.task == 'stereo':
                curr_vect = torch.clamp(curr_vect, max=0)
            ms_pred.append([prob_map, curr_vect * 2**(self.args.downsample - l), up_feat])

            if l < self.levels - 1:
                up_curr_vect = 2 * F.interpolate(
                    curr_vect,
                    scale_factor=2,
                    mode='bilinear',
                    align_corners=True)

        ms_prob = [l[0] for l in ms_pred]
        ms_vect = [l[1] for l in ms_pred]

        outputs = {}
        if self.training:
            outputs['ms_prob'] = ms_prob
            outputs['ms_pred'] = ms_vect
            outputs['corr_range'] = self.args.corr_range
            outputs['downsample'] = self.args.downsample
            outputs['flows'] = F.interpolate(ms_vect[-1], scale_factor=4, mode='bilinear', align_corners=False)[:, None]
        else:
            outputs['flows'] = F.interpolate(ms_vect[-1], scale_factor=4, mode='bilinear', align_corners=False)[:, None]
        return outputs


class HD3(BaseHD3):
    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-chairs-0d46c9fd.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-things-afcd2eb4.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-sintel-10689995.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3-kitti-6eb77dd3.ckpt'
    }

    def __init__(self, args):
        super(HD3, self).__init__(
            args, context=False)


class HD3Context(BaseHD3):
    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-chairs-d7448468.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-things-f1681f75.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-sintel-22e655e9.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/hd3_ctxt-kitti-e7d69776.ckpt'
    }

    def __init__(self, args):
        super(HD3Context, self).__init__(
            args, context=True)



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
import torch.nn.functional as tf


def _elementwise_epe(input_flow, target_flow):
    residual = target_flow - input_flow
    return torch.norm(residual, p=2, dim=1, keepdim=True)

def _elementwise_robust_epe_char(input_flow, target_flow):
    residual = target_flow - input_flow
    return torch.pow(torch.norm(residual, p=2, dim=1, keepdim=True) + 0.01, 0.4)

def _downsample2d_as(inputs, target_as):
    _, _, h, w = target_as.size()
    return tf.adaptive_avg_pool2d(inputs, [h, w])

def _upsample2d_as(inputs, target_as, mode="bilinear"):
    _, _, h, w = target_as.size()
    return tf.interpolate(inputs, [h, w], mode=mode, align_corners=True)

def f1_score(y_true, y_pred):
    return fbeta_score(y_true, y_pred, 1)

def fbeta_score(y_true, y_pred, beta, eps=1e-8):
    beta2 = beta ** 2

    y_pred = y_pred.float()
    y_true = y_true.float()

    true_positive = (y_pred * y_true).sum(dim=2).sum(dim=2)
    precision = true_positive / (y_pred.sum(dim=2).sum(dim=2) + eps)
    recall = true_positive / (y_true.sum(dim=2).sum(dim=2) + eps)

    return torch.mean(precision * recall / (precision * beta2 + recall + eps) * (1 + beta2))

def f1_score_bal_loss(y_pred, y_true):
    eps = 1e-8

    tp = -(y_true * torch.log(y_pred + eps)).sum(dim=2).sum(dim=2).sum(dim=1)
    fn = -((1 - y_true) * torch.log((1 - y_pred) + eps)).sum(dim=2).sum(dim=2).sum(dim=1)

    denom_tp = y_true.sum(dim=2).sum(dim=2).sum(dim=1) + y_pred.sum(dim=2).sum(dim=2).sum(dim=1) + eps
    denom_fn = (1 - y_true).sum(dim=2).sum(dim=2).sum(dim=1) + (1 - y_pred).sum(dim=2).sum(dim=2).sum(dim=1) + eps

    return ((tp / denom_tp).sum() + (fn / denom_fn).sum()) * y_pred.size(2) * y_pred.size(3) * 0.5


class MultiScaleEPE_FlowNet(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_FlowNet, self).__init__()
        self._args = args        
        self._batch_size = args.train_batch_size
        self._weights = [0.005, 0.01, 0.02, 0.08, 0.32]

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs = [output_dict[key] for key in ["flow2", "flow3", "flow4", "flow5", "flow6"]]

            # div_flow trick
            target = self._args.div_flow * target_dict["flows"][:, 0]

            total_loss = 0
            for i, output_i in enumerate(outputs):
                target_i = _downsample2d_as(target, output_i)
                epe_i = _elementwise_epe(output_i, target_i)
                total_loss = total_loss + self._weights[i] * epe_i.sum() / self._batch_size
                loss_dict["epe%i" % (i + 2)] = epe_i.mean()
            loss_dict["loss"] = total_loss
        else:
            output = output_dict["flow1"]
            target = target_dict["flows"][:, 0]
            epe = _elementwise_epe(output, target)
            loss_dict["epe"] = epe.mean()

        return loss_dict

class MultiScaleEPE_FlowNet_IRR(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_FlowNet_IRR, self).__init__()
        self._args = args        
        self._batch_size = args.train_batch_size
        self._weights = [0.005, 0.01, 0.02, 0.08, 0.32]
        self._num_iters = args.num_iters

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs_flo = [output_dict[key] for key in ["flow2", "flow3", "flow4", "flow5", "flow6"]]

            # div_flow trick
            target_f = self._args.div_flow * target_dict["flows"][:, 0]

            total_loss = 0
            for ii, output_ii in enumerate(outputs_flo):
                target_f_ii = _downsample2d_as(target_f, output_ii[0])
                for jj, output_ii_jj in enumerate(output_ii):
                    epe_f_ii = _elementwise_epe(output_ii_jj, target_f_ii)
                    total_loss = total_loss + self._weights[ii] * epe_f_ii.sum()
                    loss_dict["epe%i" % (ii + 2)] = epe_f_ii.mean()
            loss_dict["loss"] = total_loss / self._batch_size / self._num_iters

        else:
            output = output_dict["flow1"]
            target_f = target_dict["flows"][:, 0]
            epe_f = _elementwise_epe(target_f, output)
            loss_dict["epe"] = epe_f.mean()

        return loss_dict

class MultiScaleEPE_FlowNet_IRR_Bi(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_FlowNet_IRR_Bi, self).__init__()
        self._args = args        
        self._batch_size = args.train_batch_size
        self._weights = [0.005, 0.01, 0.02, 0.08, 0.32]
        self._num_iters = args.num_iters

    def forward(self, output_dict, target_dict):

        loss_dict = {}

        if self.training:
            outputs_flo = [output_dict[key] for key in ["flow2", "flow3", "flow4", "flow5", "flow6"]]

            # div_flow trick
            target_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_b = self._args.div_flow * target_dict["flows_b"][:, 0]

            total_loss = 0
            for ii, output_ii in enumerate(outputs_flo):
                target_f_ii = _downsample2d_as(target_f, output_ii[0][0])
                target_b_ii = _downsample2d_as(target_b, output_ii[0][1])
                for jj, output_ii_jj in enumerate(output_ii):
                    epe_f_ii = _elementwise_epe(output_ii_jj[0], target_f_ii)
                    epe_b_ii = _elementwise_epe(output_ii_jj[1], target_b_ii)
                    total_loss = total_loss + self._weights[ii] * (epe_f_ii.sum() + epe_b_ii.sum())
                    loss_dict["epe%i" % (ii + 2)] = (epe_f_ii.mean() + epe_b_ii.mean()) / 2
            loss_dict["loss"] = total_loss / self._batch_size / self._num_iters / 2
        else:
            epe_f = _elementwise_epe(output_dict["flow1"], target_dict["flows"][:, 0])
            loss_dict["epe"] = epe_f.mean()

        return loss_dict

class MultiScaleEPE_FlowNet_IRR_Occ(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_FlowNet_IRR_Occ, self).__init__()
        self._args = args        
        self._batch_size = args.train_batch_size
        self._weights = [0.005, 0.01, 0.02, 0.08, 0.32]
        self._num_iters = args.num_iters

        self.f1_score_bal_loss = f1_score_bal_loss
        self.occ_activ = nn.Sigmoid()
        
    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs_flo = [output_dict[key] for key in ["flow2", "flow3", "flow4", "flow5", "flow6"]]
            outputs_occ = [output_dict[key] for key in ["occ2", "occ3", "occ4", "occ5", "occ6"]]

            # div_flow trick
            target = self._args.div_flow * target_dict["flows"][:, 0]
            target_occ = target_dict["occs"][:, 0]

            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(outputs_flo):
                target_ii = _downsample2d_as(target, output_ii[0])
                for jj, output_ii_jj in enumerate(output_ii):
                    flow_loss = flow_loss + self._weights[ii] * _elementwise_epe(output_ii_jj, target_ii).sum()

            for ii, output_ii in enumerate(outputs_occ):
                target_occ_f = _downsample2d_as(target_occ, output_ii[0])
                for jj, output_ii_jj in enumerate(output_ii):
                    occ_loss = occ_loss + self._weights[ii] * self.f1_score_bal_loss(self.occ_activ(output_ii_jj), target_occ_f)

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size / self._num_iters
            loss_dict["occ_loss"] = occ_loss / self._batch_size / self._num_iters
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size / self._num_iters

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flow1"], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ1"])))

        return loss_dict

class MultiScaleEPE_FlowNet_IRR_Bi_Occ(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_FlowNet_IRR_Bi_Occ, self).__init__()
        self._args = args        
        self._batch_size = args.train_batch_size
        self._weights = [0.005, 0.01, 0.02, 0.08, 0.32]
        self._num_iters = args.num_iters

        self.f1_score_bal_loss = f1_score_bal_loss
        self.occ_activ = nn.Sigmoid()

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs_flo = [output_dict[key] for key in ["flow2", "flow3", "flow4", "flow5", "flow6"]]
            outputs_occ = [output_dict[key] for key in ["occ2", "occ3", "occ4", "occ5", "occ6"]]

            # div_flow trick
            target_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_b = self._args.div_flow * target_dict["flows_b"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]
            target_occ_b = target_dict["occs_b"][:, 0]

            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(outputs_flo):
                target_f_ii = _downsample2d_as(target_f, output_ii[0][0])
                target_b_ii = _downsample2d_as(target_b, output_ii[0][1])
                for jj, output_ii_jj in enumerate(output_ii):
                    epe_f_ii = _elementwise_epe(output_ii_jj[0], target_f_ii)
                    epe_b_ii = _elementwise_epe(output_ii_jj[1], target_b_ii)
                    flow_loss = flow_loss + self._weights[ii] * (epe_f_ii.sum() + epe_b_ii.sum()) * 0.5

            for ii, output_ii in enumerate(outputs_occ):
                target_occ_f = _downsample2d_as(target_occ_f, output_ii[0][0])
                target_occ_b = _downsample2d_as(target_occ_b, output_ii[0][1])
                for jj, output_ii_jj in enumerate(output_ii):
                    output_occ_f = self.occ_activ(output_ii_jj[0])
                    output_occ_b = self.occ_activ(output_ii_jj[1])
                    bce_f_ii = self.f1_score_bal_loss(output_occ_f, target_occ_f)
                    bce_b_ii = self.f1_score_bal_loss(output_occ_b, target_occ_b)
                    occ_loss = occ_loss + self._weights[ii] * (bce_f_ii + bce_b_ii) * 0.5

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size / self._num_iters
            loss_dict["occ_loss"] = occ_loss / self._batch_size / self._num_iters
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size / self._num_iters
        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flow1"], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ1"])))

        return loss_dict

class MultiScaleEPE_FlowNet_IRR_Bi_Occ_upsample(nn.Module):
    def __init__(self,
                 args):
        super(MultiScaleEPE_FlowNet_IRR_Bi_Occ_upsample, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size        
        self._weights = [0.0003125, 0.00125, 0.005, 0.01, 0.02, 0.08, 0.32]
        
        self.occ_activ = nn.Sigmoid()
        self.f1_score_bal_loss = f1_score_bal_loss

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs_flo = [output_dict[key] for key in ["flow", "flow1", "flow2", "flow3", "flow4", "flow5", "flow6"]]
            outputs_occ = [output_dict[key] for key in ["occ", "occ1", "occ2", "occ3", "occ4", "occ5", "occ6"]]

            # div_flow trick
            target_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_b = self._args.div_flow * target_dict["flows_b"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]
            target_occ_b = target_dict["occs_b"][:, 0]

            num_iters = len(outputs_flo[0])
            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(outputs_flo):
                target_f_ii = _downsample2d_as(target_f, output_ii[0][0])
                target_b_ii = _downsample2d_as(target_b, output_ii[0][1])
                for jj, output_ii_jj in enumerate(output_ii):
                    epe_f_ii = _elementwise_epe(output_ii_jj[0], target_f_ii)
                    epe_b_ii = _elementwise_epe(output_ii_jj[1], target_b_ii)
                    flow_loss = flow_loss + self._weights[ii] * (epe_f_ii.sum() + epe_b_ii.sum()) * 0.5

            for ii, output_ii in enumerate(outputs_occ):
                target_occ_f = _downsample2d_as(target_occ_f, output_ii[0][0])
                target_occ_b = _downsample2d_as(target_occ_b, output_ii[0][1])
                for jj, output_ii_jj in enumerate(output_ii):
                    output_occ_f = self.occ_activ(output_ii_jj[0])
                    output_occ_b = self.occ_activ(output_ii_jj[1])
                    bce_f_ii = self.f1_score_bal_loss(output_occ_f, target_occ_f)
                    bce_b_ii = self.f1_score_bal_loss(output_occ_b, target_occ_b)
                    occ_loss = occ_loss + self._weights[ii] * (bce_f_ii + bce_b_ii) * 0.5

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size / num_iters
            loss_dict["occ_loss"] = occ_loss / self._batch_size / num_iters
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size / num_iters
        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict



class MultiScaleEPE_PWC(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005]

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs = output_dict['flow_preds']

            # div_flow trick
            target = self._args.div_flow * target_dict["flows"][:, 0]

            total_loss = 0
            for ii, output_ii in enumerate(outputs):
                loss_ii = _elementwise_epe(output_ii, _downsample2d_as(target, output_ii)).sum()
                total_loss = total_loss + self._weights[ii] * loss_ii
            loss_dict["loss"] = total_loss / self._batch_size

        else:
            epe = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0])
            loss_dict["epe"] = epe.mean()

        return loss_dict

class MultiScaleEPE_PWC_Bi(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005]

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            outputs = output_dict['flow_preds']

            # div_flow trick
            target_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_b = self._args.div_flow * target_dict["flows_b"][:, 0]

            total_loss = 0
            for i, output_i in enumerate(outputs):
                epe_i_f = _elementwise_epe(output_i[0], _downsample2d_as(target_f, output_i[0]))
                epe_i_b = _elementwise_epe(output_i[1], _downsample2d_as(target_b, output_i[1]))
                total_loss = total_loss + self._weights[i] * (epe_i_f.sum() + epe_i_b.sum())
            loss_dict["loss"] = total_loss / (2 * self._batch_size)
        else:
            epe = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0])
            loss_dict["epe"] = epe.mean()

        return loss_dict

class MultiScaleEPE_PWC_Occ(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Occ, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005]

        self.occ_activ = nn.Sigmoid()
        self.f1_score_bal_loss = f1_score_bal_loss

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo = self._args.div_flow * target_dict["flows"][:, 0]
            target_occ = target_dict["occs"][:, 0]

            flow_loss = 0
            occ_loss = 0

            for i, output_i in enumerate(output_flo):
                flow_loss = flow_loss + self._weights[i] * _elementwise_epe(output_i, _downsample2d_as(target_flo, output_i)).sum()

            for i, output_i in enumerate(output_occ):
                output_occ = self.occ_activ(output_i)
                occ_loss = occ_loss + self._weights[i] * self.f1_score_bal_loss(output_occ, _downsample2d_as(target_occ, output_occ))

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["occ_loss"] = occ_loss / self._batch_size
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict

class MultiScaleEPE_PWC_Bi_Occ(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size        
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005]

        self.occ_activ = nn.Sigmoid()
        self.f1_score_bal_loss = f1_score_bal_loss

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_flo_b = self._args.div_flow * target_dict["flows_b"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]
            target_occ_b = target_dict["occs_b"][:, 0]

            # bchw
            flow_loss = 0
            occ_loss = 0

            for i, output_i in enumerate(output_flo):
                flow_loss = flow_loss + self._weights[i] * _elementwise_epe(output_i[0], _downsample2d_as(target_flo_f, output_i[0])).sum()
                flow_loss = flow_loss + self._weights[i] * _elementwise_epe(output_i[1], _downsample2d_as(target_flo_b, output_i[1])).sum()

            for i, output_i in enumerate(output_occ):
                output_occ_f = self.occ_activ(output_i[0])
                output_occ_b = self.occ_activ(output_i[1])
                occ_loss = occ_loss + self._weights[i] * self.f1_score_bal_loss(output_occ_f, _downsample2d_as(target_occ_f, output_occ_f))
                occ_loss = occ_loss + self._weights[i] * self.f1_score_bal_loss(output_occ_b, _downsample2d_as(target_occ_b, output_occ_b))

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / (2 * self._batch_size)
            loss_dict["occ_loss"] = occ_loss / (2 * self._batch_size) 
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / (2 * self._batch_size)

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict

class MultiScaleEPE_PWC_Bi_Occ_upsample(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005, 0.00125, 0.0003125]

        self.occ_activ = nn.Sigmoid()
        self.f1_score_bal_loss = f1_score_bal_loss

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_flo_b = self._args.div_flow * target_dict["flows_b"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]
            target_occ_b = target_dict["occs_b"][:, 0]

            # bchw
            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(output_flo):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    loss_ii = loss_ii + _elementwise_epe(output_ii[2 * jj], _downsample2d_as(target_flo_f, output_ii[2 * jj])).sum()
                    loss_ii = loss_ii + _elementwise_epe(output_ii[2 * jj + 1], _downsample2d_as(target_flo_b, output_ii[2 * jj + 1])).sum()
                flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii)

            for ii, output_ii in enumerate(output_occ):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    output_occ_f = self.occ_activ(output_ii[2 * jj])
                    output_occ_b = self.occ_activ(output_ii[2 * jj + 1])
                    loss_ii = loss_ii + self.f1_score_bal_loss(output_occ_f, _downsample2d_as(target_occ_f, output_occ_f))
                    loss_ii = loss_ii + self.f1_score_bal_loss(output_occ_b, _downsample2d_as(target_occ_b, output_occ_b))
                occ_loss = occ_loss + self._weights[ii] * loss_ii / len(output_ii)

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["occ_loss"] = occ_loss / self._batch_size
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict

class MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size        
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005, 0.00125, 0.0003125]

        self.occ_activ = nn.Sigmoid()
        self.occ_loss_bce = nn.BCELoss(reduction='sum')

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]

            # bchw
            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(output_flo):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    loss_ii = loss_ii + _elementwise_robust_epe_char(output_ii[2 * jj], _downsample2d_as(target_flo_f, output_ii[2 * jj])).sum()
                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
                flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii) * 2

            for ii, output_ii in enumerate(output_occ):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    output_occ_f = self.occ_activ(output_ii[2 * jj])
                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
                    loss_ii = loss_ii + self.occ_loss_bce(output_occ_f, _downsample2d_as(target_occ_f, output_occ_f))
                occ_loss = occ_loss + self._weights[ii] * loss_ii / len(output_ii) * 2

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if f_loss > o_loss:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["occ_loss"] = occ_loss / self._batch_size
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict

class MultiScaleEPE_PWC_Bi_Occ_upsample_KITTI(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample_KITTI, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.001, 0.001, 0.001, 0.002, 0.004, 0.004, 0.004]

        self.occ_activ = nn.Sigmoid()
        
    def forward(self, output_dict, target_dict):
        loss_dict = {}

        valid_mask = target_dict["input_valid"]
        b, _, h, w = target_dict["flows"][:, 0].size()

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]

            # bchw
            flow_loss = 0

            for ii, output_ii in enumerate(output_flo):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    valid_epe = _elementwise_robust_epe_char(_upsample2d_as(output_ii[2 * jj], target_flo_f), target_flo_f) * valid_mask

                    for bb in range(0, b):
                        valid_epe[bb, ...][valid_mask[bb, ...] == 0] = valid_epe[bb, ...][valid_mask[bb, ...] == 0].detach()
                        norm_const = h * w / (valid_mask[bb, ...].sum())
                        loss_ii = loss_ii + valid_epe[bb, ...][valid_mask[bb, ...] != 0].sum() * norm_const

                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
                flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii) * 2

            for ii, output_ii in enumerate(output_occ):
                for jj in range(0, len(output_ii) // 2):
                    output_ii[2 * jj] = output_ii[2 * jj].detach()
                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["loss"] = flow_loss / self._batch_size

        else:
            flow_gt_mag = torch.norm(target_dict["flows"][:, 0], p=2, dim=1, keepdim=True) + 1e-8
            flow_epe = _elementwise_epe(output_dict["flows"][:, 0], target_dict["flows"][:, 0]) * valid_mask

            epe_per_image = (flow_epe.view(b, -1).sum(1)) / (valid_mask.view(b, -1).sum(1))
            loss_dict["epe"] = epe_per_image.mean()

            outlier_epe = (flow_epe > 3).float() * ((flow_epe / flow_gt_mag) > 0.05).float() * valid_mask
            outlier_per_image = (outlier_epe.view(b, -1).sum(1)) / (valid_mask.view(b, -1).sum(1))
            loss_dict["outlier"] = outlier_per_image.mean()

        return loss_dict


"""Implement a callback to log images."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Union

try:
    from neptune.new.types import File as NeptuneFile
except ImportError:
    NeptuneFile = None
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.loggers.base import LoggerCollection
from pytorch_lightning.loggers.comet import CometLogger
from pytorch_lightning.loggers.neptune import NeptuneLogger
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger
from pytorch_lightning.loggers.wandb import WandbLogger
from pytorch_lightning.trainer.trainer import Trainer
import torch
import torch.nn.functional as F
from torchvision.utils import make_grid
try:
    import wandb
except ImportError:
    wandb = None

from ptlflow.models.base_model.base_model import BaseModel
from ptlflow.utils import flow_utils
from ptlflow.utils.utils import config_logging

config_logging()


class LoggerCallback(Callback):
    """Callback to collect and log images during training and validation.

    For each dataloader, num_images samples will be collected. The samples are collected by trying to retrieve from both
    inputs and outputs tensors whose keys match the values provided in log_keys.

    num_images samples are uniformly sampled from the whole dataloader.
    """

    def __init__(
        self,
        num_images: int = 5,
        image_size: Tuple[int, int] = (200, 400),
        log_keys: Sequence[str] = ('images', 'flows', 'occs', 'mbs', 'confs'),
        epe_clip: float = 5.0
    ) -> None:
        """Initialize LoggerCallback.

        Parameters
        ----------
        num_images : int, default 5
            Number of images to log during one epoch.
        image_size : Tuple[int, int], default (200, 400)
            The size of the stored images.
        log_keys : Sequence[str], default ('images', 'flows', 'occs', 'mbs', 'confs')
            The keys to use to collect the images from the inputs and outputs of the model. If a key is not found, it is
            ignored.
        epe_clip : float, default 5.0
            The maximum EPE value that is shown on EPE image. All EPE values above this will be clipped.
        """
        super().__init__()

        self.num_images = num_images
        self.image_size = image_size
        self.log_keys = log_keys
        self.epe_clip = epe_clip

        self.train_collect_img_idx = []
        self.train_images = {}

        self.val_dataloader_names = []
        self.val_collect_image_idx = {}
        self.val_images = {}

    def log_image(
        self,
        title: str,
        image: torch.Tensor,
        pl_module: BaseModel
    ) -> None:
        """Log the image in all of the pl_module loggers.

        Note, however, that not all loggers may be able to log images.

        Parameters
        ----------
        title : str
            A title for the image.
        image : torch.Tensor
            The image to log. It must be a 3D tensor CHW (typically C=3).
        pl_module : BaseModel
            An instance of the optical flow model to get the logger from.
        """
        image_npy = image.permute(1, 2, 0).numpy()

        logger_collection = pl_module.logger
        if logger_collection is not None:
            if not isinstance(logger_collection, LoggerCollection):
                logger_collection = LoggerCollection([logger_collection])

            for logger in logger_collection:
                if isinstance(logger, CometLogger):
                    logger.experiment.log_image(image_npy, name=title)
                elif isinstance(logger, NeptuneLogger):
                    logger.experiment[title].log(NeptuneFile.as_image(image))
                elif isinstance(logger, TensorBoardLogger):
                    logger.experiment.add_image(title, image, pl_module.global_step)
                elif isinstance(logger, WandbLogger) and wandb is not None:
                    title_wb = title.replace('/', '-')
                    image_wb = wandb.Image(image_npy)
                    logger.experiment.log({title_wb: image_wb})

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int
    ) -> None:
        """Store one image to be logged, if the current batch_idx is in the log selection group.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Dict[str, torch.Tensor]
            The outputs of the current training batch.
        batch : Dict[str, torch.Tensor]
            The inputs of the current training batch.
        batch_idx : int
            The counter value of the current batch.
        dataloader_idx : int
            The index number of the current dataloader.
        """
        if batch_idx in self.train_collect_img_idx:
            self._append_images(self.train_images, pl_module.last_inputs, pl_module.last_predictions)

    def on_train_epoch_start(
        self,
        trainer: Trainer,
        pl_module: BaseModel
    ) -> None:
        """Reset the training log params and accumulators.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        self.train_images = {}
        collect_idx = np.unique(np.linspace(
            0, self._compute_max_range(pl_module.train_dataloader_length, pl_module.args.limit_train_batches),
            self.num_images, dtype=np.int32))
        self.train_collect_img_idx = collect_idx

    def on_train_epoch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel,
        outputs: Any = None  # This arg does not exist anymore, but it is kept here for compatibility
    ) -> None:
        """Log the images accumulated during the training.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Any
            Outputs of the training epoch.
        """
        img_grid = self._make_image_grid(self.train_images)
        self.log_image('train', img_grid, pl_module)

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel,
        outputs: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: int
    ) -> None:
        """Store one image to be logged, if the current batch_idx is in the log selection group.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        outputs : Dict[str, torch.Tensor]
            The outputs of the current validation batch.
        batch : Dict[str, torch.Tensor]
            The inputs of the current validation batch.
        batch_idx : int
            The counter value of the current batch.
        dataloader_idx : int
            The index number of the current dataloader.
        """
        dl_name = self.val_dataloader_names[dataloader_idx]
        if batch_idx in self.val_collect_image_idx[dl_name]:
            self._append_images(self.val_images[dl_name], pl_module.last_inputs, pl_module.last_predictions)

    def on_validation_epoch_start(
        self,
        trainer: Trainer,
        pl_module: BaseModel
    ) -> None:
        """Reset the validation log params and accumulators.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        self.val_dataloader_names = pl_module.val_dataloader_names
        for dl_name in self.val_dataloader_names:
            self.val_images[dl_name] = {}

        for dname, dlen in zip(pl_module.val_dataloader_names, pl_module.val_dataloader_lengths):
            collect_idx = np.unique(np.linspace(
                0, self._compute_max_range(dlen, pl_module.args.limit_val_batches), self.num_images, dtype=np.int32))
            self.val_collect_image_idx[dname] = collect_idx

    def on_validation_epoch_end(
        self,
        trainer: Trainer,
        pl_module: BaseModel
    ) -> None:
        """Log the images accumulated during the validation.

        Parameters
        ----------
        trainer : Trainer
            An instance of the PyTorch Lightning trainer.
        pl_module : BaseModel
            An instance of the optical flow model.
        """
        for dl_name, dl_images in self.val_images.items():
            img_grid = self._make_image_grid(dl_images)
            self.log_image(f'val/{dl_name}', img_grid, pl_module)

    def _add_title(
        self,
        image: torch.Tensor,
        img_title: str
    ) -> torch.Tensor:
        """Add a title to an image.

        Parameters
        ----------
        image : torch.Tensor
            The image where the title will be added.
        img_title : str
            The title to be added.

        Returns
        -------
        torch.Tensor
            The input image with the title superposed on it.
        """
        size = min(image.shape[1:3])
        image = (255*image.permute(1, 2, 0).numpy()).astype(np.uint8)
        image = Image.fromarray(image)

        this_dir = Path(__file__).resolve().parent
        title_font = ImageFont.truetype(str(this_dir / 'RobotoMono-Regular.ttf'), size//10)

        draw = ImageDraw.Draw(image)
        bb = (size//25, size//25, size//25+len(img_title)*size//15, size//25+size//8)
        draw.rectangle(bb, fill='black')
        draw.text((size//20, size//30), img_title, (237, 230, 211), font=title_font)

        image = np.array(image)
        image = torch.from_numpy(image.transpose(2, 0, 1)).float() / 255

        return image

    def _append_images(  # noqa: C901
        self,
        images: Dict[str, List[torch.Tensor]],
        inputs: Dict[str, torch.Tensor],
        preds: Dict[str, torch.Tensor]
    ) -> None:
        """Append samples to the images accumulator.

        Parameters
        ----------
        images : Dict[str, List[torch.Tensor]]
            The accumulator where the samples will be appended to.
        inputs : Dict[str, torch.Tensor]
            The inputs of the model.
        preds : Dict[str, torch.Tensor]
            The outrputs of the model.
        """
        for k in self.log_keys:
            log_names = []
            log_sources = []
            if k in inputs or (k == 'confs' and k in preds):
                log_names.append(f'i_{k}')
                log_sources.append(inputs)
            if k in preds:
                log_names.append(f'o_{k}')
                log_sources.append(preds)

                if k == 'flows':
                    log_names.append(f'epe<{self.epe_clip:.1f}')
                    log_sources.append(None)

            for name, source in zip(log_names, log_sources):
                if images.get(name) is None:
                    images[name] = []

                if name == 'i_confs':
                    img = self._compute_confidence_gt(preds['flows'], inputs['flows'])
                elif name.startswith('epe'):
                    epe = torch.norm(preds[k] - inputs[k], p=2, dim=2, keepdim=True)
                    img = torch.clamp(epe, 0, self.epe_clip) / self.epe_clip
                    if inputs.get('valids') is not None:
                        img[inputs['valids'] < 0.5] = 0
                else:
                    img = source[k]

                img = img[:1, 0].detach().cpu()
                img = F.interpolate(img, self.image_size)
                img = img[0]

                if 'images' in name:
                    img = img.flip([0])  # BGR to RGB
                elif 'flows' in name:
                    img = flow_utils.flow_to_rgb(img)

                images[name].append(img)

    def _compute_confidence_gt(
        self,
        pred_flows: torch.Tensor,
        target_flows: torch.Tensor
    ) -> torch.Tensor:
        """Compute a confidence score for the flow predictions.

        This score was proposed in https://arxiv.org/abs/2007.09319.

        Parameters
        ----------
        pred_flows : torch.Tensor
            The predicted optical flow.
        target_flows : torch.Tensor
            The groundtruth optical flow.

        Returns
        -------
        torch.Tensor
            The confidence score for each pixel of the input.
        """
        conf_gt = torch.exp(-torch.pow(pred_flows - target_flows, 2).sum(dim=2, keepdim=True))
        return conf_gt

    def _compute_max_range(
        self,
        dataloader_length: int,
        limit_batches: Union[float, int]
    ) -> int:
        """Find the maximum number of samples that will be drawn from a dataloader.

        Parameters
        ----------
        dataloader_length : int
            Total size of the dataloader.
        limit_batches : Union[float, int]
            A value that may decrease the samples in the dataloader. See --limit_val_batches or --limit_train_batches from
            PyTorch Lightning for more information.

        Returns
        -------
        int
            The maximum number of samples that will be drawn from the dataloader.
        """
        if isinstance(limit_batches, int):
            max_range = limit_batches - 1
        else:
            max_range = int(limit_batches * dataloader_length) - 1
        return max_range

    def _make_image_grid(
        self,
        dl_images: Dict[str, List[torch.Tensor]]
    ) -> torch.Tensor:
        """Transform a bunch of images into a single one by adding them to a grid.

        Parameters
        ----------
        dl_images : Dict[str, List[torch.Tensor]]
            Lists of images, each identified by a title name.

        Returns
        -------
        torch.Tensor
            A single 3D tensor image 3HW.
        """
        imgs = []
        for img_label, img_list in dl_images.items():
            for j, im in enumerate(img_list):
                if len(im.shape) == 2:
                    im = im[None]

                if im.shape[0] == 1:
                    im = im.repeat(3, 1, 1)

                if j == 0:
                    im = self._add_title(im, img_label)

                imgs.append(im)

        grid = make_grid(imgs, len(imgs)//len(dl_images))
        return grid



"""Helper to facilitate the use of inputs and outputs of the models."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from ptlflow.data.flow_transforms import Compose, Resize, ToTensor
from ptlflow.models.base_model.base_model import BaseModel
from ptlflow.utils.utils import InputPadder


class IOAdapter(object):
    """Handle the inputs and outputs of optical flow models."""

    def __init__(
        self,
        model: BaseModel,
        input_size: Tuple[int, int],
        target_size: Optional[Tuple[int, int]] = None
    ) -> None:
        """Initialize IOAdapter.

        Parameters
        ----------
        model : BaseModel
            An instance of optical flow mode that will be used for estimation.
        input_size : Tuple[int, int]
            The shape of the original inputs, must be a tuple with at least two elements. It is assumed that the last two
            elements are (height, with).
        target_size : Optional[Tuple[int, int]], optional
            If provided, the inputs will be resized to target_size. target_size is defined as a tuple (height, with).
        """
        self.output_stride = model.output_stride
        self.target_size = target_size
        self.transform = Compose([ToTensor(), Resize(target_size)])
        self.padder = InputPadder(input_size, model.output_stride)

    def prepare_inputs(
        self,
        images: Union[np.ndarray, List[np.ndarray]],
        flows: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        **kwargs: Union[np.ndarray, List[np.ndarray]]
    ) -> Dict[str, torch.Tensor]:
        """Transform numpy inputs into the input format of theoptical flow models.

        This basically consists on tranform the numpy arrays into torch tensors, and then putting them into a dict.

        Parameters
        ----------
        images : Union[np.ndarray, List[np.ndarray]]
            One or more images to use to estimate the optical flow. Typically, it will be a least with two images in the
            HWC format.
        flows : Optional[Union[np.ndarray, List[np.ndarray]]], optional
            One or more groundtruth optical flow, which can be used for validation. Typically it will be an array HWC.
        kwargs : Union[np.ndarray, List[np.ndarray]]
            Any other array inputs can be provided as keyworded arguments. This function will create an entry in the input dict
            for each keyworded array given.

        Returns
        -------
        Dict[str, torch.Tensor]
            The inputs converted to the input format of the optical flow models.
        """
        inputs = {'images': images, 'flows': flows}
        inputs.update(kwargs)
        keys_to_remove = []
        for k, v in inputs.items():
            if v is None or len(v) == 0:
                keys_to_remove.append(k)
        for k in keys_to_remove:
            del inputs[k]

        inputs = self.transform(inputs)
        inputs = {k: self.padder.pad(v) for k, v in inputs.items()}
        inputs = {k: v.unsqueeze(0) for k, v in inputs.items()}
        return inputs

    def unpad(
        self,
        outputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Remove padding to restore the original shapes.

        The inputs and outputs obtained from prepare_inputs() will usually have some additional padding which was added to make
        the input shape divisible by the model output stride. This function can be used to remove that padding.

        Typically, this function should be used on both the inputs and outputs of the model after the model generated the
        predictions.

        Parameters
        ----------
        outputs : Dict[str, torch.Tensor]
            It can be either the inputs or the outputs of optical flow models. All tensors will have the extra padding removed.

        Returns
        -------
        Dict[str, torch.Tensor]
            The same tensors, with the padding removed.

        See Also
        --------
        prepare_inputs
        """
        outputs = {k: self.padder.unpad(v) if isinstance(v, torch.Tensor) else v for k, v in outputs.items()}
        return outputs



from argparse import Namespace
from pathlib import Path

import torch
import torch.nn as nn
from torch.nn import init

from .submodules import *
from .flownet_base import FlowNetBase

class FlowNetC(FlowNetBase):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetc-things-cc8ac7fd.ckpt'
    }

    def __init__(self, args: Namespace):
        super(FlowNetC,self).__init__(args)

        self.args = args

        self.rgb_max = 1

        self.conv1   = conv(self.args.batch_norm,   3,   64, kernel_size=7, stride=2)
        self.conv2   = conv(self.args.batch_norm,  64,  128, kernel_size=5, stride=2)
        self.conv3   = conv(self.args.batch_norm, 128,  256, kernel_size=5, stride=2)
        self.conv_redir  = conv(self.args.batch_norm, 256,   32, kernel_size=1, stride=1)

        self.corr = correlate #Correlation(pad_size=20, kernel_size=1, max_displacement=20, stride1=1, stride2=2, corr_multiply=1)

        self.corr_activation = nn.LeakyReLU(0.1,inplace=True)
        self.conv3_1 = conv(self.args.batch_norm, 473,  256)
        self.conv4   = conv(self.args.batch_norm, 256,  512, stride=2)
        self.conv4_1 = conv(self.args.batch_norm, 512,  512)
        self.conv5   = conv(self.args.batch_norm, 512,  512, stride=2)
        self.conv5_1 = conv(self.args.batch_norm, 512,  512)
        self.conv6   = conv(self.args.batch_norm, 512, 1024, stride=2)
        self.conv6_1 = conv(self.args.batch_norm,1024, 1024)

        self.deconv5 = deconv(1024,512)
        self.deconv4 = deconv(1026,256)
        self.deconv3 = deconv(770,128)
        self.deconv2 = deconv(386,64)

        self.predict_flow6 = predict_flow(1024)
        self.predict_flow5 = predict_flow(1026)
        self.predict_flow4 = predict_flow(770)
        self.predict_flow3 = predict_flow(386)
        self.predict_flow2 = predict_flow(194)

        self.upsampled_flow6_to_5 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow5_to_4 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow4_to_3 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)
        self.upsampled_flow3_to_2 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear', align_corners=False)

    def normalize(self, im):
        im = im - 0.5
        im = im / 0.5
        return im

    def forward(self, inputs):
        x = inputs['images']
        x1 = x[:, 0]
        x2 = x[:, 1]

        x1 = self.normalize(x1)
        x2 = self.normalize(x2)
        # FlownetC top input stream
        out_conv1a = self.conv1(x1)
        out_conv2a = self.conv2(out_conv1a)
        out_conv3a = self.conv3(out_conv2a)

        # FlownetC bottom input stream
        out_conv1b = self.conv1(x2)

        out_conv2b = self.conv2(out_conv1b)
        out_conv3b = self.conv3(out_conv2b)

        # Merge streams
        out_corr = self.corr(out_conv3a, out_conv3b) # False
        out_corr = self.corr_activation(out_corr)

        # Redirect top input stream and concatenate
        out_conv_redir = self.conv_redir(out_conv3a)

        in_conv3_1 = torch.cat((out_conv_redir, out_corr), 1)

        # Merged conv layers
        out_conv3_1 = self.conv3_1(in_conv3_1)

        out_conv4 = self.conv4_1(self.conv4(out_conv3_1))

        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)

        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)

        flow5       = self.predict_flow5(concat5)
        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)

        flow4       = self.predict_flow4(concat4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        concat3 = torch.cat((out_conv3_1,out_deconv3,flow4_up),1)

        flow3       = self.predict_flow3(concat3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)
        concat2 = torch.cat((out_conv2a,out_deconv2,flow3_up),1)

        flow2 = self.predict_flow2(concat2)

        outputs = {}

        if self.training:
            outputs['flow_preds'] = [flow2,flow3,flow4,flow5,flow6]
            outputs['flows'] = self.upsample1(flow2*self.args.div_flow)[:, None]
        else:
            outputs['flows'] = self.upsample1(flow2*self.args.div_flow)[:, None]

        return outputs



"""Function taken and adapted from https://github.com/ppliuboy/SelFlow/blob/master/flowlib.py."""

# Original license below

# MIT License

# Copyright (c) 2019 Pengpeng Liu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

import numpy as np


def write_pfm(output_path, flow, scale=1):
    with open(output_path, 'wb') as file:
        if flow.dtype.name != 'float32':
            raise TypeError('flow dtype must be float32.')
        if not (len(flow.shape) == 3 and flow.shape[2] == 2):
            raise ValueError('flow must have H x W x 2 shape.')

        file.write(b'PF\n')
        file.write(b'%d %d\n' % (flow.shape[1], flow.shape[0]))

        endian = flow.dtype.byteorder
        if endian == '<' or endian == '=' and sys.byteorder == 'little':
            scale = -scale

        file.write(b'%f\n' % scale)

        invalid = np.isnan(flow[..., 0]) | np.isnan(flow[..., 1])
        flow = np.dstack([flow, invalid.astype(np.float32)])
        flow = np.flipud(flow)
        flow.tofile(file)
